module Channel where


import Control.Concurrent as CC 
import Control.Exception
import Control.Monad
import MyQSem as MS
import System.IO.Unsafe
import Random
import List
import Maybe

data Channel i o = Channel { writeC :: [Port i], readC :: [Port i], fC :: i -> o, canI :: CC.QSem,
                             writeSyncList :: MVar [([String], QSemN, Maybe i)], 
                             readSyncList :: MVar [([String], QSemN, Maybe i)], 
                             writeRep :: MVar Int, readRep :: MVar Int,
                             writeSync :: MVar Int, readSync :: MVar Int,
                             onlyOne :: MVar Bool
                           }

data Port a = Prt { idp :: PortId, hole :: MVar (Stream a) }

type PortId = String

type Stream a = MVar (Item a) 

data Item a = Item a (Stream a) deriving Eq 


instance Show a => Show (Item a)
  where
    show (Item a as) = "Item " ++ show a ++ " " ++ show_a as
        where
            show_a a = unsafePerformIO $ do 
                            mv <- readMVar a
                            return $ show mv               
  

show_ (Prt idp h) = read_x idp h
        where
            read_x idp h = do 
                              x <- readMVar h
                              y <- show_x x
                              return $ "Prt { idp = " ++ idp ++ ", hole: " ++ y ++ " }, "
            show_x x = do 
                        is <- isEmptyMVar x
                        if is
                            then do 
                                return ""
                            else do
                                Item val next <- readMVar x
                                rest <- show_x next
                                return $ show val ++ " " ++ rest

testIs = 
    do
    x <- newMVar 'a'
    y <- isEmptyMVar x
    putStrLn $ show y


showCh (Channel w r _ _ _ _ _ _ _ _ _) = do
    showw <- mapM show_ w
    showr <- mapM show_ r
    return $ "Channel { write = " ++ concat showw ++ " read = " ++ concat showr ++ " }"




readCh :: (Eq i, Show o) => Channel i o -> PortId -> IO [o]
readCh ch@(Channel writeC readC f canI _ acts _ rep _ max _) idr 
  =
  block $ do 
                CC.signalQSem canI 
                rep' <- readMVar rep
                max' <- readMVar max
                if max' /= 1
                    then do
                        verActR' idr acts [] rep' 
                        x <- readAction readC idr f rep' []
                        return x
                    else do
                        x <- readAction readC idr f rep' []
                        return x

readAction :: [Port i] -> String -> (i -> o) -> Int -> [o] -> IO [o]
readAction _ _ _ 0 l = return l
readAction readC idr f rep l = 
    block $ do
        cts <- takeMVarR readC idr
        x <- takeMVar cts
        Item val new <- takeMVar x
        putMVarR readC idr new
        readAction readC idr f (rep-1) ((f val):l)
   
verActR' idr acts other max = 
    block $ do 
    ls <- takeMVar acts
    if length ls == 0
        then do 
            putMVar acts ls
      else do 
            let (ids, wo', xx):r = ls
            if idr `elem` ids 
                then do 
                    let list = delete idr ids
                    if list == []
                        then do 
                            let y = other ++ r
                            putMVar acts y 
                            signalQSemN wo' max 
                        else do 
                                let x = other ++ (list, wo', xx):r
                                putMVar acts x
                                waitQSemN wo' 1 
                else do 
                    putMVar acts r
                    verActR' idr acts ((ids, wo', xx):other) max 



writeCh (Channel w r _ canI acts _ rep _ max _ oo) val idr = 
    block $ do
                CC.waitQSem canI 
                rep' <- readMVar rep
                max' <- readMVar max
                if max' /= 1
                    then do
                        rep'' <- verActW' idr acts [] max' val rep' oo
                        writeAction w idr val rep''
                    else do 
                        writeAction w idr val rep' 
    where
    mapList (f:fs) (x:xs) = (f x) : (mapList fs xs)
    mapList _ _ = []

    mapIO f [] _ = return ()
    mapIO f (x:xs) (p:ps) = do 
                            x' <- x
                            f' f x' p >> mapIO f xs ps
        where f' _ [] _ = return ()
              f' f (h:t) p = f h p >> f' f t p
       
writeAction :: [Port a] -> String -> a -> Int -> IO ()
writeAction w idr val 0 = return ()
writeAction w idr val rep = 
    block $ do 
        new_hole <- newEmptyMVar
        old_bufferl <- takeMVarR w idr
        modifyMVar_ old_bufferl (\old_hole -> do
                                                putMVar old_hole (Item val new_hole)
                                                return new_hole)
        writeAction w idr val (rep-1)



verActW' idr acts other max val rep oo = 
    block $ do 
    oo' <- readMVar oo
    ls <- takeMVar acts
    if length ls == 0
        then do 
            putMVar acts ls
            return rep
        else do 
            let (ids, wo', tow):r = ls
            if tow == Nothing && not oo' 
                then do
                    doS idr ids r wo' max rep acts val other oo
                else do
                    if Just val == tow
                        then do
                            doS idr ids r wo' max 0 acts val other oo
                        else do
                            putMVar acts r
                            verActW' idr acts ((ids, wo', tow):other) max val rep oo

doS idr ids r wo' maxR rep acts val other oo = 
                if idr `elem` ids
                    then do 
                        let list = delete idr ids
                        if list == [] 
                            then do 
                                let y = other ++ r
                                putMVar acts y 
                                signalQSemN wo' maxR 
                                return rep
                            else do
                                let x = other ++ (list, wo', Just val):r
                                putMVar acts x
                                waitQSemN wo' 1
                                return rep
                    else do 
                        putMVar acts r
                        verActW' idr acts ((ids, wo', Just val):other) maxR val rep oo



takeMVarR :: Monad m => [Port a] -> String -> m (MVar (Stream a))
takeMVarR [] idr = error $ "take: " ++ idr ++" that port does not exists"
takeMVarR ((Prt idr h):t) idr1 | idr == idr1 = return h
                               | otherwise = takeMVarR t idr1


putMVarR :: [Port a] -> String -> Stream a -> IO ()
putMVarR [] idr1 _ = error $ "put: " ++ idr1 ++" that port does not exists"
putMVarR ((Prt idr h):t) idr1 new | idr == idr1 = putMVar h new
                                  | otherwise = putMVarR t idr1 new



getAllPortIds [] = []
getAllPortIds ((Prt idr _):t) = idr:(getAllPortIds t)




barrier ps = do
    mvars <- replicateM (length ps) newEmptyMVar
    mapM (\(mv, p) -> forkIO (p >>= putMVar mv)) (zip mvars ps)
    mapM takeMVar mvars

    


