module Connectors where

import Channel
import Control.Concurrent
import Control.Monad

hook :: (Eq o, Show o, Eq o, Show k) 
    => 
    Channel o o -> Channel o k -> [(String, String)] -> IO (Channel o k)
hook ch1@(Channel w1 r1 a1 b1 c1 d1 e1 f1 g1 h1 i1 ) 
     ch2@(Channel w2 r2 a2 b2 c2 d2 e2 f2 g2 h2 i2 ) l 
    =
    if l == []
        then 
            error "there are no ports to connect" 
        else 
            let readPrts = map fst l
                writePrts = map snd l
            in if not (isAllIn readPrts (map idp r1)) || not (isAllIn writePrts (map idp w2))
                then
                  error "incompatible ports"
                else do 
                    forkIO $ readWrite readPrts writePrts
                    let w1' = getNotUsedPorts writePrts $ w1++w2
                    let r2' = getNotUsedPorts readPrts $ r1++r2
                    return $ Channel w1' r2' (a2.a1) b1 c1 d2 e1 f2 g1 h2 i1 
    where
        readWrite readPrts writePrts = do
                    runInBoundThread $ do
                        xpto <- barrier $ mapM__ (readCh ch1) readPrts
                        let xpto2 = (map_ (map (writeCh_ ch2) writePrts)) xpto
                        barrier xpto2
                        return () 
                    readWrite readPrts writePrts

writeCh_ x p t = writeCh x t p

mapM__ f [] = []
mapM__ f (h:t) = (f h):(mapM__ f t)
                                     
map_ [] _ = []
map_ (h:t) (x:xs) = do 
                            let xpto = barrier $ mapM__ h x
                            let rs = map_ t xs
                            xpto:rs

isAllIn l1 l2 = and $ map (`elem` l2) l1

getNotUsedPorts _ [] = []
getNotUsedPorts l (pt@(Prt id h):t) = if id `elem` l
                                            then getNotUsedPorts l t
                                            else pt:(getNotUsedPorts l t)




