module MyQSem where


import Control.Concurrent


newtype QSem = QSem (MVar (Int, Int, [MVar ()]))

instance Eq MyQSem.QSem
  where
    (==) (QSem x) (QSem y) = x == y

newQSem :: Int -> Int -> IO MyQSem.QSem
newQSem init max = do
   sem <- newMVar (init, max, [])
   return (QSem sem)

waitQSem :: MyQSem.QSem -> IO ()
waitQSem (QSem sem) = do
   (avail, max, blocked) <- takeMVar sem  
   if avail > 0 
     then do
            putMVar sem (avail-1, max, [])
     else do
       block <- newEmptyMVar
       putMVar sem (0, max, blocked++[block])
       takeMVar block

signalQSem :: MyQSem.QSem -> IO ()
signalQSem (QSem sem) = do
    (avail, max, blocked) <- takeMVar sem
    if avail < max
        then
            case blocked of
                [] -> do
                         putMVar sem (avail+1, max, [])

                (block:blocked') -> do
                    	   putMVar sem (0, max, blocked')
                    	   putMVar block ()
        else do 
               block <- newEmptyMVar
               putMVar sem (avail+1, max, [block])
               takeMVar block


getMaxQSem :: MyQSem.QSem -> IO Int
getMaxQSem (QSem q) = do 
    (_, m, _) <- readMVar q
    return m

getCurQSem :: MyQSem.QSem -> IO Int
getCurQSem (QSem q) = do 
    (c, _, _) <- readMVar q
    return c

getListLenQSem :: MyQSem.QSem -> IO Int
getListLenQSem (QSem q) = do 
    (_, _, l) <- readMVar q
    return $ length l

